#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "qwsim.h"

/*Complex number functions*/


complex c_sum(complex z1, complex z2){
/*Receives two complex numbers z1 and z2 and returns r = z_1 + z_2*/
	complex r;
	r.re = z1.re + z2.re;
	r.im = z1.im + z2.im;

	return r;
}

complex c_mult(complex z1, complex z2){
/*Receives two complex numbers z1 and z2 and returns r = z_1*z_2 */
	double a,b,c,d;
	complex r;
	a = z1.re;
	b = z1.im;
	c = z2.re;
	d = z2.im;
	r.re = a*c-b*d;
	r.im = b*c+a*d;
	return r;
}

double c_angle(complex z){
/*Receives complex number z and returns arctan(Re(z)/Im(z))*/
	double r = 0.0;
	if(z.re > 0.0){
		r = atan(z.im/z.re);
	}
	if(z.re < 0.0 && z.im >= 0.0){
		r = atan(z.im/z.re) + M_PI;
	}
	if(z.re < 0.0 && z.im < 0.0){
		r = atan(z.im/z.re) - M_PI;
	}
	if(z.re == 0.0 && z.im > 0.0){
		r = M_PI/2.0;
	}
	if(z.re == 0.0 && z.im < 0.0){
		r = -M_PI/2.0;
	}
    
	return r;
}

complex c_div(complex z1, complex z2){
/*Receives complex numbers z1 and z2 and returns z1/z2*/
	double a,b,c,d;
	complex r;
	a = z1.re;
	b = z1.im;
	c = z2.re;
	d = z2.im;

	r.re = (a*c+b*d)/(c*c + d*d);
	r.im = (b*c-a*d)/(c*c + d*d);

	return r;
}

complex c_conj(complex c){
/*Receives complex number z and retunrs its complex conjugate*/
	complex z;

	z.re = c.re;
	z.im = -c.im;

	return z;
}


double c_abs_sq(complex z){
/*Receives complex number z and outputs it squared modulus*/
double a = z.re;
double b = z.im;

return a*a+b*b;

}

/*complex matrix functions*/

double vector_norm_sq(c_matrix a){
/*Receives vector a and returns squared norm of a*/
	int i;
	double r = 0.0;
	for(i=0; i < a.rows; i++){
		r = r + c_abs_sq(M(a,i,0));
	}
	return r;
}

void new_matrix(c_matrix *r,int rows,int cols){
/*Initializes matrix r with "rowns" rows and "cols" columns with all entries equal to 0.0*/
	int i;
	r->rows=rows;
	r->cols=cols;
	r->t=(complex *)malloc(sizeof(complex)*cols*rows);
	for(i=0;i< rows*cols; i++){
		r->t[i].re = 0.0;
		r->t[i].im = 0.0;
	}

}

void reset_matrix(c_matrix *r){
/*Receives a matrix r and resets its entries to 0.0*/
	int i;
	for(i=0;i< (r->rows)*(r->cols); i++){
		r->t[i].re = 0.0;
		r->t[i].im = 0.0;
	}
	
}

void erase_matrix(c_matrix *a){
/*Erases matrix a*/
	free(a->t);
	a->t = NULL;
}

void print_matrix(c_matrix t) {
/*Prints matrix t on screen*/
	int i,j;
	for(i=0;i<t.rows;i++) {
		printf("| ");
		for(j=0;j<t.cols;j++)
			printf("%f +i%f ",M(t,i,j).re,M(t,i,j).im);
		printf("|\n");
	}
	printf("\n");
}

void matrix_mult(c_matrix r,c_matrix a,c_matrix b) {
/*multiplies matrix a with b and storees result on r*/
	int m,n,k;
	for(m=0;m<a.rows;m++){
		for(k=0;k<b.cols;k++) {
			M(r,m,k).re = 0.0;
			M(r,m,k).im = 0.0;
			for(n=0;n<a.cols;n++){
				M(r,m,k)=c_sum(M(r,m,k),c_mult(M(a,m,n),M(b,n,k)));
			}
		}
	}
}

void matrix_sum(c_matrix r, c_matrix a, c_matrix b) {
/*Sums matrix a with b and storees result on r*/
	int i,j;
	for(i=0;i<a.rows;i++){
		for(j=0;j<b.cols;j++) {
			M(r,i,j)=c_sum(M(a,i,j),M(b,i,j));
		}
	}
}

void c_times_matrix(c_matrix r, complex c, c_matrix a) {
/*multiplies constant c with matrix a and the result is stored on r*/
	int i,j;
	for(i=0;i<r.rows;i++){
		for(j=0;j<r.cols;j++) {
			M(r,i,j)=c_mult(c,M(a,i,j));
		}
	}
}


void matrix_conj(c_matrix r,c_matrix a){
/*stores in matrix r the complex conjugate of the entries of matrix a */
	int i,j;

	for(i = 0; i < a.cols; i++){
		for(j=0; j < a.rows; j ++){
			M(r,i,j) = c_conj(M(a,i,j));
		}
	}
}

void matrix_transpose(c_matrix r, c_matrix a){
/* stores in matrix r the transpose of matrix a */
	int i,j;

	for(i=0;i<a.rows;i++){
		for(j=0;j<a.cols;j++){
			M(r,j,i) = M(a,i,j);
		}
	}
}

void matrix_adjoint(c_matrix *r,c_matrix a){
/*stores in matrix r the adjoint of matrix a*/
	int i,j;
	for(i=0;i<r->rows;i++){
		for(j=0;j<r->cols;j++){
			M((*r),i,j) = c_conj(M(a,j,i));
		}
	}
}

void position_trace_1D(state *s,int t){
/*Given state |s> this function determines the partial trace over the coin space*/
	int c1,c2,i_1,j_1, i_2, j_2;

	for(c1=0; c1 < Edges; c1++)
		for(c2=0; c2 < Edges; c2++){
			M(rho_c,c1,c2).re = 0.0;
			M(rho_c,c1,c2).im = 0.0;
	}
	for(c1=0; c1 < Edges; c1++)
		for(c2=0; c2 < Edges; c2++)
			for(i_1=-t; i_1 <= t; i_1++)
	M(rho_c,c1,c2) = c_sum(M(rho_c,c1,c2), c_mult(M(s[biject1D(i_1)].direction,c1,0), c_conj(M(s[biject1D(i_1)].direction,c2,0))));
}




void identity(c_matrix I){
/*Initializes matrix I with entries of the identity matrix */
	int i;

	for(i=0; i < I.cols; i++){
		M(I,i,i).re = 1.0;
		M(I,i,i).im = 0.0;
	}
}

void copy_matrix(c_matrix r, c_matrix a){
/*Copies entries of matrix a into matrix r*/
	int i,j;

	for(i = 0; i < a.rows; i++){
		for(j=0; j < a.cols; j ++){
			M(r,i,j) = M(a,i,j);
		}
	}
}

c_matrix I_diag_B(c_matrix a, int dim){
/*Returns a block diagonal matrix r of dimension "dim" where the first block is the diagonal matrix and the second is matrix a */
	c_matrix r;
	int i,j,ir,jr;
	i = -1;

	new_matrix(&r,dim,dim);
	identity(r);
	for(ir= dim-a.rows; ir < dim; ir++){
		i++;
		j=-1;
		for(jr=dim-a.cols; jr < dim; jr++){
			j++;
			M(r,ir,jr) = M(a,i,j);
		}
	}
	return r;
}

c_matrix copy_sub_matrix(c_matrix a,int *lines, int *cols){
/*Copies the entries of matrix a from lines[0] to lines[1] and cols[0] to cols[1] into matrix r*/
	c_matrix r;
	int i,j,ir,jr;
	ir = -1;

	new_matrix(&r,1+lines[1]-lines[0],1+cols[1]-cols[0]);
	for(i=lines[0]; i <= lines[1]; i++){
		ir++;
		jr=-1;
		for(j=cols[0]; j <= cols[1]; j++){
			jr++;
			M(r,ir,jr) = M(a,i,j);
		}
	}
	return r;
}

c_matrix QR_decomposition(c_matrix A){
/*Returns matrix Q from QR decompositionn of A, A = QR*/
	int i,lins[2],cols[2];
	c_matrix x, P_i,B_i, Qt, Q_aux,PA, PA_aux,Q;

	lins[1] = A.cols-1;
	new_matrix(&Qt,A.rows, A.cols);
	new_matrix(&Q_aux,A.rows, A.cols);
	new_matrix(&PA, A.rows, A.cols);
	new_matrix(&PA_aux,PA.rows, PA.cols);
	new_matrix(&Q,A.rows, A.cols);
	copy_matrix(PA,A);
	identity(Q);

	for(i=0; i < A.cols; i++){
		lins[0] = i;
		cols[0] = i;
		cols[1] = i;

		x = copy_sub_matrix(PA,lins,cols);
		new_matrix(&B_i, 1+lins[1]-lins[0],1+ lins[1]-lins[0]);

		Householder_matrix(B_i,x);
		P_i = I_diag_B(B_i,A.cols);

		matrix_mult(PA_aux,P_i,PA);
		copy_matrix(PA, PA_aux);

		matrix_mult(Q_aux,P_i,Q);
		copy_matrix(Q,Q_aux);

		erase_matrix(&B_i);
		erase_matrix(&P_i);
		erase_matrix(&x);
	}
	erase_matrix(&PA);
	erase_matrix(&PA_aux);
	erase_matrix(&Q_aux);

	erase_matrix(&Qt);
	return Q;
}

c_matrix matrix_P(c_matrix a, c_matrix b){
/*Given matrix A, B, returns R = [A 0, 0 B]*/
	int i, j, ik, jk, n;
	c_matrix r;

	n = a.cols+b.cols;
	new_matrix(&r,n,n);

	for(i=0;i < a.rows; i++){
		for(j=0;j < a.cols; j++){
			M(r,i,j) = M(a,i,j);
		}
	}
	ik= a.rows;
	for(i=0;i < b.rows; i++){
		jk=a.cols;
		for(j=0;j < b.cols; j++){
				M(r,ik,jk) = M(b,i,j);
				jk++;
			}
		ik++;
	}
	return r;
}

void Householder_matrix(c_matrix P, c_matrix x){
/*Computes the Householder matrix for QR decomposition*/
	c_matrix v, v_ad;
	int i;
	double gamma = c_angle(M(x,0,0));
	double v_norm_sq;
	double x_norm = sqrt(vector_norm_sq(x));
	complex e_gamma;
	complex aux, aux_2;

	e_gamma.re = cos(gamma);

	e_gamma.im = sin(gamma);


	new_matrix(&v,x.rows,x.cols);
	new_matrix(&v_ad, x.cols, x.rows);

	copy_matrix(v,x);

	aux_2.re = x_norm;
	aux_2.im = 0.0;
	aux = c_mult(e_gamma, aux_2);

	M(v,0,0) = c_sum(M(v,0,0) , aux);

	v_norm_sq = vector_norm_sq(v);
	matrix_adjoint(&v_ad, v);

	matrix_mult(P,v,v_ad);

	aux.re = 0.0;
	aux.im = 0.0;
	if(v_norm_sq != 0.0)
		aux.re = -2.0/v_norm_sq;

	c_times_matrix(P,aux, P);
	aux.re = 1.0;
	for(i=0; i < P.rows; i++){
		M(P,i,i) = c_sum(M(P,i,i), aux);
	}

	erase_matrix(&v_ad);
	erase_matrix(&v);
}

double cost_function(c_matrix A){
/*returns the sum of squared absolute values of matrix A's diagonal matrix*/
	int i;
	double r = 0.0;

	for(i=0;i < A.cols;i++){
		r = r + c_abs_sq(M(A,i,i));
	}
	return r;
}

void QR_algorithm2(c_matrix A,c_matrix R ,c_matrix Qn){
/*Same as QR, but changes the values of R and Qn such that A = QRn */
	int k;
	int tm = 0;
	int tf = 0;
	complex c;
	c_matrix Qn_aux,Q,Qt,Ak, Ak_1,Aux, Aux2;
	c.re = -1.0;
	c.im = 0.0;

	new_matrix(&Ak,A.cols,A.cols);
	tm++;
	new_matrix(&Ak_1,A.cols,A.cols);
	tm++;
	new_matrix(&Aux,A.cols,A.cols);
	tm++;
	new_matrix(&Aux2, A.cols, A.cols);
	tm++;
	new_matrix(&Q, A.cols, A.cols);
	tm++;
	new_matrix(&Qn_aux, A.cols, A.rows);
	tm++;
	identity(Qn_aux);
	copy_matrix(Qn,Qn_aux);
	copy_matrix(Ak_1,A);

	for(k=1;k<=500;k++){

		Qt = QR_decomposition(Ak_1);
	tm++;
		matrix_adjoint(&Q,Qt);
		matrix_mult(Qn_aux, Qt,Qn);
		copy_matrix(Qn,Qn_aux);
		matrix_mult(R,Qt,Ak_1);
		matrix_mult(Ak,R,Q);

		c_times_matrix(Aux, c, Ak_1);
		copy_matrix(Ak_1,Ak);

		matrix_sum(Aux2, Aux, Ak);

		erase_matrix(&Qt);
		tf++;
	}
	erase_matrix(&Ak_1);
	tf++;
	erase_matrix(&Aux);
	tf++;
	erase_matrix(&Aux2);
	tf++;
	erase_matrix(&Ak);
	tf++;
	erase_matrix(&Qn_aux);
	tf++;
	erase_matrix(&Q);
	tf++;
	for(k=0;k<R.cols; k++){
		M(R,k,k).re = - M(R,k,k).re;
		M(R,k,k).im = - M(R,k,k).im;
	}

}



c_matrix QR_algorithm(c_matrix A){
/*Returns matrix R of QR factorization A = QR*/
	int k;
	complex c;
	c_matrix Q,Qt,R,Ak, Ak_1,Aux, Aux2;
	c.re = -1.0;
	c.im = 0.0;
	new_matrix(&R,A.cols,A.cols);
	new_matrix(&Ak,A.cols,A.cols);
	new_matrix(&Ak_1,A.cols,A.cols);
	new_matrix(&Aux,A.cols,A.cols);
	new_matrix(&Aux2, A.cols, A.cols);
	new_matrix(&Q, A.cols, A.cols);
	new_matrix(&Qt,A.cols, A.cols);
	copy_matrix(Ak_1,A);

	for(k=1;k<=1000;k++){

		Qt = QR_decomposition(Ak_1);
		matrix_adjoint(&Q,Qt);
		matrix_mult(R,Qt,Ak_1);
		matrix_mult(Ak,R,Q);



		c_times_matrix(Aux, c, Ak_1);
		copy_matrix(Ak_1,Ak);


		matrix_sum(Aux2, Aux, Ak);

		erase_matrix(&Qt);
	}

	erase_matrix(&Ak_1);
	erase_matrix(&Aux);
	erase_matrix(&Aux2);
	erase_matrix(&Ak);
	erase_matrix(&Q);
	erase_matrix(&Qt);
	for(k=0;k<R.cols; k++){
		M(R,k,k).re = - M(R,k,k).re;
		M(R,k,k).im = - M(R,k,k).im;
	}
	return R;

}

complex entropy(c_matrix D){
/*Determines the von Neumann entropy/Shannon entropy of diagonal matrix D*/
	int i;
	complex S;
	double aux;
	S.re = 0.0;
	S.im = 0.0;

	for(i=0; i < D.cols; i++){
		if(c_abs_sq(M(D,i,i)) <= pow(10.0,-10.0)){
			S.re = S.re + 0.0;
			S.im = S.im + 0.0;
		}
		else
		{
		aux = sqrt(c_abs_sq(M(D,i,i)));

		S.re = S.re - aux*log(aux);
		S.im = S.im - c_angle(M(D,i,i));
		}
	}

return S;
}
/*Quantum Walk functions*/
int biject1D(int a){
	return a+N;
}

int bijinv1D(int a){
	return a-N;
}

state * initialize_state(){
/*Allocates memory for the overall quantum state*/
	state *s = (state *)malloc(sizeof(state) * Nodes);
	int i,j;

	for(i=0;i< Nodes  ; i++)
	{
		s[i].direction.cols = 1;
		s[i].direction.rows = Edges;
		s[i].direction.t=(complex *)malloc(sizeof(complex)*Edges);
		for(j=0;j<Edges; j++){
			M(s[i].direction,j,0).re = 0.0;
			M(s[i].direction,j,0).im = 0.0;
		}
	}
	return s;
}

void erase_state(state *s){
/*Erases quantum state from memory*/
	int i;

	for(i=0;i< Nodes  ; i++)
		{
			s[i].direction.cols = 0;
			s[i].direction.rows = 0;
			free(s[i].direction.t);
		}

}


void copy_state(state *s1, state*s2){
/*Copies all the entries from quantum state s1 to quantum state s2*/
int i,e,j;

	for(i=0;i<Nodes ;i++){
		for(e=0; e < Edges; e++){
			M(s1[i].direction,e,0) = M(s2[i].direction,e,0);
			M(s2[i].direction,e,0).re = 0.0;
			M(s2[i].direction,e,0).im = 0.0;

		}
	}


}

void shift1D(state *s1, state *s2, int t){
/*Performs the shift operation on quantum state s_1 and stores it on state s_2*/
	int i;

	for(i=-N;i <= N; i++){

		shift_cond[biject1D(i)][0] = biject1D(i+1);
		shift_cond[biject1D(i)][1] = 0;

		shift_cond[biject1D(i)][2] = biject1D(i-1);
		shift_cond[biject1D(i)][3] = 1;
			
	}

}

void reflect1D(){
/*sets the code -2 (reflecting coin) to positions N and -N of vector pos_coin */
	pos_coin[biject1D(N)] = -1;
	pos_coin[biject1D(-N)] = -1;


}

void absorbing_boundary(int position,int c,int t){
/*performs the absortion operation on state s1 in position "position" and coin state "c"*/
		double normalizer;
		int i,j;

		normalizer =sqrt(1.0 -  c_abs_sq(M(s1[biject1D(position)].direction,c,0)));
		
		measure_prob[t] = measure_prob[t] + c_abs_sq(M(s1[biject1D(position)].direction,c,0));

		M(s1[biject1D(position)].direction,c,0).re = 0.0;
		M(s1[biject1D(position)].direction,c,0).im = 0.0;

	for(i=0;i< Nodes  ; i++){
		for(j=0;j<Edges; j++){
			M(s1[i].direction,j,0).re = M(s1[i].direction,j,0).re/normalizer;
			M(s1[i].direction,j,0).im = M(s1[i].direction,j,0).im/normalizer;
		}
	}
}



void initial_broken_link1D(int i, int direction){
/*sets the initial broken link at position "i" on direction "direction"*/
	switch(direction){
		case 0: 

				shift_cond[biject1D(i  )][ 2*0   ] = biject1D(i);
				shift_cond[biject1D(i  )][ 2*0+1 ] = 1;

				shift_cond[biject1D(i+1)][ 2*1   ] = biject1D(i+1);
				shift_cond[biject1D(i+1)][ 2*1+1 ] = 0;
			break;
		case 1:

				shift_cond[biject1D(i )][ 2*1   ] = biject1D(i);
				shift_cond[biject1D(i )][ 2*1+1 ] = 0;

				shift_cond[biject1D(i-1)][ 2*0   ] = biject1D(i-1);
				shift_cond[biject1D(i-1)][ 2*0+1 ] = 1;

			break;

	}

}

void broken_link1D(int i, int direction){
/*sets the broken link at position "i" on direction "direction"*/
	switch(direction){
		case 0: 
				shift_cond_altered[biject1D(i  )][ 2*0   ] = biject1D(i);
				shift_cond_altered[biject1D(i  )][ 2*0+1 ] = 1;

				shift_cond_altered[biject1D(i+1)][ 2*1   ] = biject1D(i+1);
				shift_cond_altered[biject1D(i+1)][ 2*1+1 ] = 0;

			break;
		case 1:

				shift_cond_altered[biject1D(i )][ 2*1   ] = biject1D(i);
				shift_cond_altered[biject1D(i )][ 2*1+1 ] = 0;

				shift_cond_altered[biject1D(i-1)][ 2*0   ] = biject1D(i-1);
				shift_cond_altered[biject1D(i-1)][ 2*0+1 ] = 1;
			break;

	}

}

void permanent_broken_link1D(int i, int direction){
/*sets the broken link at position "i" on direction "direction"*/
	switch(direction){
		case 0: 
				shift_cond[biject1D(i  )][ 2*0   ] = biject1D(i);
				shift_cond[biject1D(i  )][ 2*0+1 ] = 1;

				shift_cond[biject1D(i+1)][ 2*1   ] = biject1D(i+1);
				shift_cond[biject1D(i+1)][ 2*1+1 ] = 0;

			break;
		case 1:

				shift_cond[biject1D(i )][ 2*1   ] = biject1D(i);
				shift_cond[biject1D(i )][ 2*1+1 ] = 0;

				shift_cond[biject1D(i-1)][ 2*0   ] = biject1D(i-1);
				shift_cond[biject1D(i-1)][ 2*0+1 ] = 1;
			break;

	}

}

void circle1D(){
/*sets the circle configuration on the line*/
		shift_cond[biject1D(N)][0] = biject1D(-N);	//next position
		shift_cond[biject1D(N)][1] = 0;			//coin state in next position


		shift_cond[biject1D(-N)][2] = biject1D(N);	//next position
		shift_cond[biject1D(-N)][3] = 1;			//coin state in next position
		

}


void step(state *s1, state *s2, int t){
/*Performs one step of the quantum walk*/
int k,i,j;

	if(varying_coin_each_step == 1){
		reset_matrix(&random_coin);
		random_matrix_1D_3(0);
		copy_matrix(randomt, random_coin);
	}
	coin1D(s1,s2, t);
	copy_shift_cond();
	link_impurities();

	for(i=-N;i <= N; i++){
		k = biject1D(i);

		M(   s2[  shift_cond_altered[k][0]  ].direction  ,  shift_cond_altered[k][1]   ,    0) = M(s1[k].direction,0,0);	
		M(   s2[  shift_cond_altered[k][2]  ].direction  ,  shift_cond_altered[k][3]   ,    0) = M(s1[k].direction,1,0);

	}

copy_state(s1,s2);
}

void initialize_shift_cond(){
/*Allocates and initializes entries of ''shift_cond''*/
	int i,j;

	shift_cond = calloc(Nodes,sizeof(int *));

	for(i=0; i<Nodes; i++){
		shift_cond[i] = calloc(2*Edges,sizeof(int));
		for(j=0;j<2*Edges;j++)
			shift_cond[i][j] = 0;
	}

}

void initialize_shift_cond_altered(){
/*Allocates and initializes entries of ''shift_cond_altered''*/
	int i,j;

	shift_cond_altered = calloc(Nodes,sizeof(int *));

	for(i=0; i<Nodes; i++){
		shift_cond_altered[i] = calloc(2*Edges,sizeof(int));
		for(j=0;j<2*Edges;j++)
			shift_cond_altered[i][j] = 0;
	}

}

void choose_coin(k){
/*Given position ''k'', the matrix random_coin will be set according to ''pos_coin_altered[k]'' */
	reset_matrix(&random_coin);
	switch(pos_coin_altered[k]){
	case -2:
		identity(random_coin);
		break;
	case -1:
		reflect_coin_1D(random_coin);
		break;
	case 0: 
		Hadamard1D();
		break;
	case 3: 
		random_matrix_1D();
		break;
	case 4:
		random_matrix_1D_2(k);
		break;
	case 5:
		copy_matrix(random_coin, randomt);
	}

}

void copy_position_coin(){
/*Copies all entries of ''pos_coin'' to ''pos_coin_altered''*/
	int i;

	for(i=0; i< Nodes; i++)
		pos_coin_altered[i] = pos_coin[i];

}

void coin1D(state *s1, state *s2, int t){
/*Applies the coin operator in each position of the line*/
	int i,j,k;
	t = N-1;
	copy_position_coin();
	coin_impurities();
	for(i=-N;i <= N; i++){
        k = biject1D(i);
			choose_coin(k);
			matrix_mult(s2[k].direction, random_coin, s1[k].direction);
	}
	copy_state(s1,s2);
}

void initialize_position_coin(){
/*Initializes ''position_coin'' with coin ''coin1''*/
	int i;
	pos_coin = (int *)malloc(sizeof(int )*Nodes);

	for(i=0; i< Nodes; i++)
		pos_coin[i] = coin1;

}

void initialize_position_coin_altered(){
    /*Initializes ''position_coin_altered'' with ''coin1''*/
	int i;
	pos_coin_altered = (int *)malloc(sizeof(int )*Nodes);

	for(i=0; i< Nodes; i++)
		pos_coin_altered[i] = coin1;

}



void plot2D(char *filename,double *xvals, double *yvals, int NUM_POINTS){
    char *aux = (char *)malloc(sizeof(char) * (strlen(filename) + 8));
    strcpy(aux,"Output/");
    strcat(aux, filename);
    FILE * temp = fopen(aux, "w");


    int i;

    	for (i=0; i < NUM_POINTS; i++)
    	{
    		fprintf(temp, "%lf %lf \n", xvals[i], yvals[i]); //Write the data to a temporary file
    	}

    	fclose(temp);
    free(aux);

}

void plot3D(char *filename, double *zvals, int NUM_POINTS){


    char *aux = (char *)malloc(sizeof(char) * (strlen(filename) + 8));
    strcpy(aux,"Output/");
    strcat(aux, filename);
    FILE * temp = fopen(aux, "w");


    if(temp == NULL)
    	printf("\n não abriu ficheiro nenhum \n");

    int i;
    	for (i=0; i < NUM_POINTS; i++)
    		fprintf(temp, "%lf %lf %lf \n", x[i], y[i], zvals[i]);

    	fclose(temp);

    free(aux);
}

void Hadamard1D(){
/*Sets ''random_coin'' with Hadamard matrix*/
	M(random_coin,0,0).re = 1.0/sqrt(2.0);
	M(random_coin,0,1).re = 1.0/sqrt(2.0);
	M(random_coin,1,0).re = 1.0/sqrt(2.0);
	M(random_coin,1,1).re = -1.0/sqrt(2.0);
}


double random_0_1(){
/*Returns a random number from 0 to 1*/
	double a,b,c;

	a = rand();
	b = RAND_MAX;
	c = a/b;
	return c;

}

double random_0_Nodes(){
/*Returns a random integer number between 0 and the ''Nodes''*/
	double a,b,c;

	a = rand();
	b = RAND_MAX;
	c = a/b*Nodes;
	return c;

}

double random_01_1(){
/*Returns a random number from -1 to 1*/
	double a,b,c,d;

	a = rand();
	b = RAND_MAX;
	c = a/b;
	a = rand();
	d = a/b;
	if(d < 0.5)
		c = -c;

	return c;
}



void random_matrix_1D(){
/*Returns a random 2x2 unitary matrix*/
	double theta, zeta, xi;
	complex e_xi, e_zeta, cos_theta, sin_theta,c;

	xi = random_0_1()*M_PI/2.0;
	zeta = random_0_1()*M_PI/2.0;
	theta = random_0_1()*M_PI/2.0;

	e_xi.re = cos(xi);
	e_xi.im = sin(xi);
	e_zeta.re = cos(zeta);
	e_zeta.im = sin(zeta);
	cos_theta.re = cos(theta);
	cos_theta.im = 0.0;
	sin_theta.re = sin(theta);
	sin_theta.im = 0.0;

	c.re = -1.0;
	c.im = 0.0;

	M(random_coin,0,0) = c_mult(e_xi, cos_theta);
	M(random_coin,0,1) = c_mult(e_zeta, sin_theta);
	e_zeta.im = sin(-zeta);
	M(random_coin,1,0) = c_mult(e_zeta, sin_theta);
	e_xi.im = sin(-xi);
	M(random_coin,1,1) = c_mult(e_xi, c_mult(c,cos_theta));

}


void random_matrix_1D_2(int pos){
/*Returns a random 2x2 unitary matrix*/
	double theta, zeta, xi;
	complex e_xi, e_zeta, cos_theta, sin_theta,c;

	theta = range[pos][0] + random_0_1()*(range[pos][1] - range[pos][0]);
	zeta = range[pos][2] + random_0_1()*(range[pos][3] - range[pos][2]);
	xi = range[pos][4] + random_0_1()*(range[pos][5] - range[pos][4]);;

	e_xi.re = cos(xi);
	e_xi.im = sin(xi);
	e_zeta.re = cos(zeta);
	e_zeta.im = sin(zeta);
	cos_theta.re = cos(theta);
	cos_theta.im = 0.0;
	sin_theta.re = sin(theta);
	sin_theta.im = 0.0;

	c.re = -1.0;
	c.im = 0.0;

	M(random_coin,0,0) = c_mult(e_xi, cos_theta);
	M(random_coin,0,1) = c_mult(e_zeta, sin_theta);
	e_zeta.im = sin(-zeta);
	M(random_coin,1,0) = c_mult(e_zeta, sin_theta);
	e_xi.im = sin(-xi);
	M(random_coin,1,1) = c_mult(e_xi, c_mult(c,cos_theta));

}

void random_matrix_1D_3(int pos){
/*Returns a random 2x2 unitary matrix*/
	double theta, zeta, xi;
	complex e_xi, e_zeta, cos_theta, sin_theta,c;

	theta = range_fixed[pos][0] + random_0_1()*(range_fixed[pos][1] - range_fixed[pos][0]);
	zeta = range_fixed[pos][2] + random_0_1()*(range_fixed[pos][3] - range_fixed[pos][2]);
	xi = range_fixed[pos][4] + random_0_1()*(range_fixed[pos][5] - range_fixed[pos][4]);;

	e_xi.re = cos(xi);
	e_xi.im = sin(xi);
	e_zeta.re = cos(zeta);
	e_zeta.im = sin(zeta);
	cos_theta.re = cos(theta);
	cos_theta.im = 0.0;
	sin_theta.re = sin(theta);
	sin_theta.im = 0.0;

	c.re = -1.0;
	c.im = 0.0;

	M(random_coin,0,0) = c_mult(e_xi, cos_theta);
	M(random_coin,0,1) = c_mult(e_zeta, sin_theta);
	e_zeta.im = sin(-zeta);
	M(random_coin,1,0) = c_mult(e_zeta, sin_theta);
	e_xi.im = sin(-xi);
	M(random_coin,1,1) = c_mult(e_xi, c_mult(c,cos_theta));

}

void mean_value_x(state *s){
/*Determines the particle's mean value position */
	int i,c,x_;
	double aux;

	mean_x[t_counter-1] = 0.0;
	for(i=0;i< Nodes; i++){
		x_ = bijinv1D(i);
		for(c=0; c < Edges; c++){
			aux = (double) x_;
			mean_x[t_counter-1] = mean_x[t_counter-1] + aux * c_abs_sq(M(s[i].direction,c,0));
		}
		
	}
}


void copy_shift_cond(){
/*Copies ''shift_cond'' into ''shift_cond_altered''*/
	int i,j;

	for(i=0;i< Nodes; i++)
		for(j=0;j< 2*Edges; j++){
			shift_cond_altered[i][j] = shift_cond[i][j];
		}
}

void link_impurities(){
/*Randomly choose broken link according to ''brokenlink_imp''*/
	int k=0;
	int i,j,c,c2, x_,y_, coin_state1, coin_state2;


	for(i=-N+1;i<N; i++){
				
			if(random_0_1() < index_brokenlink_imp){
				broken_link1D(i, 0);

			}
		}
	

}



void coin_impurities(){
/*Randomly chooses positions for random coins with respect to ''C_imp = index_coin_imp * Nodes''*/
	int i, x_,y_;
	x_ = N+1;
	for(i=0;i<C_imp; i++){
		while(abs(x_) > N)
			x_ = random_01_1()*N;
		pos_coin_altered[biject1D(x_)] = 3;
		x_ = N+1;
		y_ = N+1;
	}

}

void standard_deviation_x(state *s){
/*Determines the standard deviation of of particle's position*/
	int i,c,x_;
	double aux1, aux2;

	for(i=0;i< Nodes; i++){
		x_ = bijinv1D(i);
		aux1 = (double) x_;
		xy_cov[t_counter-1] = xy_cov[t_counter-1] + pow((aux1-mean_x[t_counter-1]),2)*P[i];
	}
	xy_cov[t_counter-1] = sqrt(xy_cov[t_counter-1]);

}

void initialize_global_variables(){
/*Initializes all the global variables and structures*/
	int i;

	Nodes1D = 2*N + 1;

	Nodes = Nodes1D;
	N_imp = index_brokenlink_imp * Nodes * Edges;
	C_imp = index_coin_imp*Nodes;

	new_matrix(&random_coin,Edges,Edges);
	new_matrix(&randomt, Edges, Edges);
	new_matrix(&rho_c,Edges, Edges);
	P_av = (double *)malloc(sizeof(double) * Nodes);
	P = (double *)malloc(sizeof(double) * Nodes);
	E_v = (double *)malloc(sizeof(double) * steps);
	E_v_p = (double *)malloc(sizeof(double) * steps);
	mean_x = (double *)malloc(sizeof(double) * steps);
	//xy_cov = (double *)malloc(sizeof(double ) * steps);
	x = (double *)malloc(sizeof(double ) * Nodes);
	y = (double *)malloc(sizeof(double ) * Nodes);
	xy_cov = (double *)malloc(sizeof(double ) * steps);
	measure_prob = (double *)malloc(sizeof(double ) * steps);

	initialize_shift_cond();
	initialize_shift_cond_altered();
	shift1D(s1,s2,t_counter);
	initialize_position_coin();
	initialize_position_coin_altered();

	for(i=0;i< Nodes; i++){
		P_av[i] = 0.0;
		P[i] = 0.0;
	}
	for(i=0;i< steps; i++){
		E_v[i] = 0.0;
		E_v_p[i] = 0.0;
		mean_x[i] = 0.0;
		xy_cov[i] = 0.0;
		measure_prob[i] = 0.0;
	}

	for(i=0;i< Nodes1D; i++){
		x[i]=0;
		y[i]=0;
	}

	s1 = initialize_state();
	s2 = initialize_state();


}

void plots(){
/*Plots in GNUplot the quantities determined in each step of the quantum walk*/
	int i,j;
	double r;
	double k1[Nodes];
	
	for(i=0; i < Nodes; i++)
		P_av[i] = P_av[i]/((double) steps);
	
	for(i=0;i < Nodes; i++){
		r = bijinv1D(i);
		x[i] = (double) r;
		
	}
	
	plot2D("Probability_Distribution\0",x,P, Nodes);
	
	plot2D("Average_Probability_Distribution\0",x,P_av, Nodes);
	
	
	
	double *k = (double *)malloc(sizeof(double ) * steps);
	//double k[steps];
	

	for(i=0 ; i<steps ; i++)
		k[i] = (double) i;
	

	plot2D("Von_Neumann_entropy\0",k,E_v,steps);
	

	plot2D("Shannon_entropy\0",k,E_v_p,steps);
	

	plot2D("mean_x\0",k,mean_x,steps);
	
	plot2D("standard_deviation\0",k, xy_cov, steps);
    
	
	printf("\n test = %lf\n",prob_test(s1));
	
	plot2D("meaasure_prob\0",k, measure_prob, steps);
	free(k);

}

void sum_prop_pos(state *s){
/*Determines the entries of ''P'', probability of the particle beeing at each position*/
	int i,j;
	for(i=0;i < Nodes; i++){
		P[i] = 0.0;
		for(j=0;j < Edges;j++)
			P[i] = P[i] + c_abs_sq(M(s[i].direction,j,0));
	}
}

double prob_test(state *s){
/*Sums all the entries of ''P''*/
	int i,j;
	double p=0.0;
	for(i=0;i < Nodes; i++)
		for(j=0;j < Edges;j++)
			p = p + c_abs_sq(M(s[i].direction,j,0));

	return p;
}

void P_av_actualization(){

	int i;
	for(i=0;i< Nodes; i++)
		P_av[i] = P_av[i] + P[i];
}
complex sum_diag(c_matrix A){
/*Returns the sums of the diagonal of matrix A */
	int i;
	complex r;
         
	r.re = 0.0;
	r.im = 0.0;

	for(i=0; i < A.cols; i++){
		r.re = M(A,i,i).re + r.re;
		r.im = M(A,i,i).im + r.im;
	}
return r;
}

void initialize_absorb(){
/*Initializes ''absorb''*/
	int i;
	absorb = (int **)malloc(sizeof(int *) * dim_absorb);
	for(i=0; i < dim_absorb; i++){
		absorb[i] = (int *)malloc(sizeof(int ) * 2);
		absorb[i][0] = 0; absorb[i][1] = 0;
	}
}

void absorbing_boundary_selected(int t){
/*Applies absorbing_boundary for every predefined absorbing points.*/
int i;
for(i=0; i < dim_absorb; i++)
	absorbing_boundary(absorb[i][0], absorb[i][1], t);
}

void qw_on_plane(state *s1, state *s2){
/*Performs the quantum walk and determines all the quantities*/
	complex c;
	int t;
    
	c_matrix diag_rho_c, Qn;
	new_matrix(&diag_rho_c,rho_c.rows,rho_c.cols);
	new_matrix(&Qn,rho_c.rows,rho_c.cols);
	for(t_counter=1;t_counter <= steps; t_counter++){
			if(t_counter > N)
				t = N;
			else
				t = t_counter;


			printf("Step = %d \n", t_counter-1);
			
			step(s1, s2,  t);
			
			absorbing_boundary_selected(t);
			
			sum_prop_pos( s1);
			
			P_av_actualization();
			
			position_trace_1D(s1,t);
			
			E_v_p[t_counter-1] = entropy(rho_c).re;
			
			c = sum_diag(rho_c);
			
			QR_algorithm2(rho_c,diag_rho_c  ,Qn);
			
			//diag_rho_c = QR_algorithm(rho_c);
			E_v[t_counter-1] = entropy(diag_rho_c).re;
			
			mean_value_x(s1);
			
			standard_deviation_x(s1);
			
	}
	erase_matrix(&diag_rho_c);
	erase_matrix(&Qn);
}



void reflect_coin_1D(c_matrix R){
/*Sets matrix ''R'' to reflecting coin*/
	M(R,0,0).re = 0.0;
	M(R,0,1).re = 1.0;
	M(R,1,0).re = 1.0;
	M(R,1,1).re = 0.0;
}

void gnu_plot2D(char *filename,double *xvals, double *yvals, int NUM_POINTS){
    
    FILE * gnuplotPipe = popen ("gnuplot -persistent", "w");
    
    int i;
    
	fprintf(gnuplotPipe,"set title '%s'\n", filename);
    	fprintf(gnuplotPipe, "plot '%s' with l\n", filename);
	pclose(gnuplotPipe);
}



void gnu_plots(){  
/*Plots in GNUplot the quantities determined in each step of the quantum walk*/
	int i,j;
	double r;
	double k1[Nodes];

	for(i=0; i < Nodes; i++)
		P_av[i] = P_av[i]/((double) steps);

	for(i=0;i < Nodes; i++){
		r = bijinv1D(i);
		x[i] = (double) r;
	}

	gnu_plot2D("Output/Probability_Distribution\0",x,P, Nodes);
	gnu_plot2D("Output/Average_Probability_Distribution\0",x,P_av, Nodes);

	//double k[steps];
	double *k = (double *)malloc(sizeof(double ) * steps);
	for(i=0 ; i<steps ; i++)
		k[i] = (double) i;

	gnu_plot2D("Output/Von_Neumann_entropy\0",k,E_v,steps);

	gnu_plot2D("Output/Shannon_entropy\0",k,E_v_p,steps);
	

	gnu_plot2D("Output/mean_x\0",k,mean_x,steps);
	gnu_plot2D("Output/standard_deviation\0",k, xy_cov, steps);
	printf("\n test = %lf\n",prob_test(s1));
    free(k);
    
}


//random_coin
