/*********This file contains all the global variables and functions/subroutines prototypes of file qwsim.c**********/

#ifndef _sim_H_
#define	_sim_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>




/******************Structures*****************************/
typedef struct complex{
double re;
double im;
}complex;

typedef struct c_matrix{
	int rows;
	int cols;
	complex *t;
}c_matrix;

/*This macro is to be used with the structure c_matrix
 *m(A,i,j) accesses the line i and column j entry of matrix A
 * */
#define M(m,x,y) m.t[x+y*m.rows]


typedef struct state{
c_matrix direction;
}state;


/*******************************Global Variables/structures********************************/

//s1 and s2 are the overall state. s1 is the main structure and the operations on s1 are stored in s2. at the end of a step, s2 is copied to s1. This will avoid erasing values when performing shift operation.
state *s1, *s2;
//Number of steps for Quantum Walk
int steps;
//Dimension of the lattice
int N;
//Number of nodes on the lattice
int Nodes;
//Maximum number of the variable x or y
int Nodes1D;
//Total number of nodes on the lattice
int Nodes2D;
//Number of edges per node (Only 4 is used)
int Edges;
//Counter for the number of steps performed
int t_counter;
//Fraction of link impurities
double index_brokenlink_imp;
int fixed_broken_link;
int fixed_coins;
//Number of links times index_brokenlink_imp
int N_imp;
int varying_coin_each_step;
//Fraction of coin impurities
double index_coin_imp;
//Number of positions times index_coin_imp
int C_imp;
//Pointer for vector that stores the code for the coin operator at each position
int *pos_coin;
//Pointer for vector that stores the code for the coin operator at each position and per step alterations
int *pos_coin_altered;
//Number of points for measurement. It sets the dimension fpr measure_ij
int N_points_measure;
//Stores the options for the mixing times.
int m_options;
//Memory check 
int N_malloc;
int N_free;
//Option for the broken link type: topology type or coin type
int bl2D;
//Choice for calculating von Neumann Entropy of coordinate x
int vn_x;
//Choice for calculating von Neumann Entropy of coordinate y
int vn_y;
//Choice for calculating von Neumann mutual information of coordinates x and y
int vn_m_info;
//Choice for calculating quantum discord
int q_discord;
//Choice for calculating upper bound for entanglement of formation
int ent_form;
//Choice for reflecting boundary on direction x
int refi;
//Choice for reflecting boundary on direction y
int refj;
//Moebius strip on the boundaries for x = +-N
int moi;
//Moebius strip on the boundaries for y = +-N
int moj;
//Choice of coin to apply on all positions of the lattice
int coin1;
//Choice for ploting output data with gnuplot
int gnu;
//Average position probability over time
double *P_av;
//Position probability
double *P;
//Pointer for vector that von Neumann entropy of coin state at each step
double *S_c;
//Pointer for vector that von Neumann entropy of variable x at each step
double *S_x;
//Pointer for vector that von Neumann entropy of variable x at each step
double *S_y;
//Pointer for vector that von Neumann mutual information of variables x and y at each step
double *Iv_xy;
//Pointer for vector that von Neumann entropy of truncated measured density matrix at each step
double *S_y_pi;
//Pointer for vector that quantum discord at each step
double *quantum_discord;
//Pointer for vector storing Shannon entropy of coin state at each step
double *H_c;
//Pointer for vector storing Shannon entropy of variable x at each step
double *H_x;
//Pointer for vector storing Shannon entropy of variable y at each step
double *H_y;
//Pointer for vector storing Shannon entropy of variables x and y at each step
double *H_xy;
//Pointer for vector storing Shannon mutual information of variables x and y at each step
double *I_xy;
//Pointer for vector storing mean value of variable x at each step
double *mean_x;
//Pointer for vector storing mean value of variable y at each step
double *mean_y;
//x and y are pointers to vectors with values from -N to N. They are used in plot3D
double *x, *y;
//Pointer for vector storing covariance of variables x and y at each step
double *xy_cov;
//Pointer for vector storing mean distance between variabes x and y at each step
double *mean_dist;
//Pointer for vector storing upper bound of entanglement of formation at each step 
double *E_f;
double ** range, ** range_fixed;
//Number of measuring points
int dim;
//Pointers for vectors storing measuring coordinates
int *i_0, *j_0;
//Coin state of the system (trace over position)
c_matrix rho_c;
//Matrix used as storage of the particular coin operator to be applied at each position and at each step 
c_matrix random_coin;
//Density matrix of system x
c_matrix X_p;
//Density matrix of system y
c_matrix Y_p;
//Truncated measured density matrix
c_matrix Y_pos;
//Stores eigenvalues of Y_pos
c_matrix Y_lambda;
c_matrix randomt;
//Pointer to a matrix that stores the topology of the lattice
int **shift_cond;
//Pointer for matrix that stores topology of the lattice and per step alterations
int **shift_cond_altered;
//Stores probability of particle in certain positions, determined by i_0 and j_0, at each step.
double *measure_prob;
//Normalization factor for state after measurement
double *norm_factor;
//Pointer for vector that stores one shot probability to hit at each step
double *one_shot_prob_hit;
//Pointer for vector that stores the average hitting time at each step
double *av_hitting_time;
//Pointer for vector that stores the first time probability to hit at each step
double *first_time_hit;
//Pointer for vector that stores the concurrent probability to hit at each step
double *concurrent_hit_time;

/*****************************************Functions and subroutines******************************************
**********************************************************************************************************/


/*Complex number operations*/
complex c_sum(complex, complex);
complex c_mult(complex, complex);
complex c_div(complex, complex);
complex c_conj(complex);
double c_abs_sq(complex);
double c_angle(complex);


/*complex matrix operations*/
double random_0_1();
double vector_norm_sq(c_matrix);
void new_matrix(c_matrix * ,int ,int );
void print_matrix(c_matrix );
void erase_matrix(c_matrix *);
void matrix_mult( c_matrix, c_matrix, c_matrix);
void matrix_sum(c_matrix ,c_matrix ,c_matrix );
void c_times_matrix(c_matrix, complex , c_matrix);
void matrix_conj(c_matrix ,c_matrix );
void matrix_transpose(c_matrix, c_matrix);
void matrix_adjoint(c_matrix *, c_matrix );
void identity(c_matrix);
void copy_matrix(c_matrix, c_matrix);
c_matrix copy_sub_matrix(c_matrix ,int *, int *);
void random_matrix_1D(c_matrix );
void random_matrix_1D_21(int, c_matrix );
void random_matrix_1D_22(int, c_matrix);
void random_matrix_2D();
void random_matrix_2D_2(int );
void random_matrix_2D_3(int );
c_matrix random_rotation_2D();
void kron(c_matrix, c_matrix, c_matrix);
void reflect_coin_1D(c_matrix );
void reflect_coin_2D(c_matrix );
void reflect_i();
void reflect_j();
int test_eigen_values(c_matrix, c_matrix);

/*Eigenvalues*/
void Householder_matrix(c_matrix, c_matrix);
c_matrix matrix_P(c_matrix , c_matrix );
c_matrix I_diag_B(c_matrix , int );
c_matrix QR_decomposition(c_matrix);
c_matrix QR_algorithm(c_matrix);
void QR_algorithm2(c_matrix, c_matrix, c_matrix);


/************************************Quantum Walk*************************************/
int biject(int, int);
int biject3D(int, int, int);
void bijinv(int* ,int);
void bijinv3D(int*, int);
state * initialize_state();
void shift(state *,state*, int );
void shift2D(state *, state *, int );
void step(state *, state *, int );
void erase_state(state *);
void display2D(state *);
void coin1D(state *, state *, c_matrix , int);
void coin2D(state *, state *, int);
void copy_state_normalized(state *,state *, int);
c_matrix position_trace_1D(state *);
void position_trace_2D(state *, int );
void coin_trace_2D(state *,int );
c_matrix Hadamard1D();
void Hadamard2D();
void Fourier2D();
void Grover2D();


/*Broken links, and graph related functions*/
void initialize_shift_cond();
void initialize_shift_cond_altered();
void reflect2D();
void moebius2D_i();
void moebius2D_j();
void broken_link2D(int , int , int);
void copy_shift_cond();
void link_impurities();
void initialize_position_coin();
void initialize_position_coin_altered();
void coin_impurities();
void permanent_broken_link2D_0(int , int , int );
void cylinder_2D_i();
void cylinder_2D_j();

/*Relevant Quantities*/
void mean_value_xy(state *);
void covariance_xy(state *);
complex entropy(c_matrix);
void rho_x();
void rho_y();
void entropy_xy();
void entropy_x();
void entropy_y();
void mutual_info_Shannon();
void mutual_info_von_neumann();

/*Most recent additions*/
void initialize_global_variables();
void sum_prop_pos(state *);
double prob_test(state *);
void P_av_actualization();
void qw_on_plane(state *s1, state *s2);


/*Measurement subroutines*/
void initialize_measure_ij();
void measure_options(int , int , int );
void entropy_truncated_measurements(int );
void quantum_discord_calc(int );

/*subroutines related to file generation and storage*/
void plot2D(char *,double *, double *, int);
void plot3D(char *, double *, int);
void plots();

/*subroutines related to GNUplot*/
void gnu_plot2D(char *, double *, double *, int);
void gnu_plot3D(char *, double *, int);
void gnu_plots();

#endif
